#!/usr/bin/env python
# -*- coding: utf-8 -*-


import os
import sys

if __name__ == "__main__":
    os.environ["DJANGO_SETTINGS_MODULE"] = "edc_consent.tests.test_settings"
    import django

    django.setup()
    from django.test.runner import DiscoverRunner

    tags = [t.split("=")[1] for t in sys.argv if t.startswith("--tag")]
    failfast = any([True for t in sys.argv if t.startswith("--failfast")])
    keepdb = any([True for t in sys.argv if t.startswith("--keepdb")])
    opts = dict(failfast=failfast, tags=tags, keepdb=keepdb)
    failures = DiscoverRunner(**opts).run_tests(["edc_consent.tests"], **opts)
    sys.exit(failures)



from dateutil.relativedelta import relativedelta
from django.contrib.sites.models import Site
from edc_constants.constants import MALE, NO, YES
from edc_utils import get_utcnow
from faker import Faker
from model_bakery.recipe import Recipe, seq

from .models import (
    SubjectConsent,
    SubjectConsentV1,
    SubjectConsentV2,
    SubjectConsentV3,
    SubjectConsentV4,
)

fake = Faker()


def get_opts():
    return dict(
        consent_datetime=get_utcnow,
        dob=get_utcnow() - relativedelta(years=25),
        first_name=fake.first_name,
        last_name=fake.last_name,
        # note, passes for model but won't pass validation in modelform clean()
        initials="AA",
        gender=MALE,
        # will raise IntegrityError if multiple made without _quantity
        identity=seq("12315678"),
        # will raise IntegrityError if multiple made without _quantity
        confirm_identity=seq("12315678"),
        identity_type="passport",
        is_dob_estimated="-",
        language="en",
        is_literate=YES,
        is_incarcerated=NO,
        study_questions=YES,
        consent_reviewed=YES,
        consent_copy=YES,
        assessment_score=YES,
        consent_signature=YES,
        site=Site.objects.get_current(),
    )


subjectconsent = Recipe(SubjectConsent, **get_opts())

subjectconsentv1 = Recipe(SubjectConsentV1, **get_opts())
subjectconsentv2 = Recipe(SubjectConsentV2, **get_opts())
subjectconsentv3 = Recipe(SubjectConsentV3, **get_opts())
subjectconsentv4 = Recipe(SubjectConsentV4, **get_opts())



from datetime import datetime
from typing import Protocol

from edc_identifier.model_mixins import NonUniqueSubjectIdentifierModelMixin
from edc_model.models import BaseUuidModel
from edc_sites.model_mixins import SiteModelMixin

from .model_mixins import ConsentModelMixin


class ConsentModelStub(Protocol):
    subject_identifier: str
    report_datetime: datetime
    ...


class ConsentLikeModel(
    SiteModelMixin, ConsentModelMixin, NonUniqueSubjectIdentifierModelMixin, BaseUuidModel
): ...



import sys
from pathlib import Path

from dateutil.relativedelta import relativedelta
from edc_test_settings.default_test_settings import DefaultTestSettings
from edc_utils import get_utcnow

app_name = "edc_consent"
base_dir = Path(__file__).absolute().parent

project_settings = DefaultTestSettings(
    calling_file=__file__,
    BASE_DIR=base_dir,
    DJANGO_CRYPTO_FIELDS_KEY_PATH=base_dir / "crypto_keys",
    GIT_DIR=base_dir.parent.parent,
    APP_NAME=app_name,
    SILENCED_SYSTEM_CHECKS=[
        "edc_consent.E001",
        "sites.E101",
        "edc_navbar.E002",
        "edc_navbar.E003",
        "edc_sites.E001",
    ],
    ETC_DIR=str(base_dir / app_name / "tests" / "etc"),
    EDC_NAVBAR_DEFAULT="edc_consent",
    EDC_PROTOCOL_STUDY_OPEN_DATETIME=get_utcnow() - relativedelta(years=1),
    EDC_PROTOCOL_STUDY_CLOSE_DATETIME=get_utcnow() + relativedelta(years=1),
    EDC_SITES_REGISTER_DEFAULT=True,
    SUBJECT_SCREENING_MODEL="consent_app.subjectscreening",
    SUBJECT_CONSENT_MODEL="consent_app.subjectconsent",
    SUBJECT_VISIT_MODEL="consent_app.subjectvisit",
    SUBJECT_VISIT_MISSED_MODEL="consent_app.subjectvisitmissed",
    SUBJECT_REQUISITION_MODEL="consent_app.subjectrequisition",
    SUBJECT_REFUSAL_MODEL="consent_app.subjectrefusal",
    INSTALLED_APPS=[
        "django.contrib.admin",
        "django.contrib.auth",
        "django.contrib.contenttypes",
        "django.contrib.sessions",
        "django.contrib.messages",
        "django.contrib.staticfiles",
        "django.contrib.sites",
        "multisite",
        "simple_history",
        "django_crypto_fields.apps.AppConfig",
        "django_revision.apps.AppConfig",
        "edc_appointment.apps.AppConfig",
        "edc_visit_schedule.apps.AppConfig",
        "edc_export.apps.AppConfig",
        "edc_crf.apps.AppConfig",
        "edc_sites.apps.AppConfig",
        "edc_form_runners.apps.AppConfig",
        "edc_timepoint.apps.AppConfig",
        "edc_data_manager.apps.AppConfig",
        "edc_device.apps.AppConfig",
        "edc_identifier.apps.AppConfig",
        "edc_action_item.apps.AppConfig",
        "edc_lab.apps.AppConfig",
        "edc_locator.apps.AppConfig",
        "edc_metadata.apps.AppConfig",
        "edc_notification.apps.AppConfig",
        "edc_protocol.apps.AppConfig",
        "edc_registration.apps.AppConfig",
        "edc_visit_tracking.apps.AppConfig",
        "edc_consent.apps.AppConfig",
        "edc_auth.apps.AppConfig",
        "consent_app.apps.AppConfig",
        "edc_appconfig.apps.AppConfig",
    ],
    add_dashboard_middleware=True,
).settings


for k, v in project_settings.items():
    setattr(sys.modules[__name__], k, v)



from .exceptions import SiteConsentError
from .site_consents import site_consents

navbar_codenames = ["edc_consent.nav_consent"]

navbar_tuples = []
for codename in navbar_codenames:
    navbar_tuples.append((codename, f"Can access {codename.split('.')[1]}"))

consent_codenames = []
try:
    cdefs = site_consents.get_consent_definitions()
except SiteConsentError:
    # error will be broadcast in system checks
    pass
else:
    models = [cdef.proxy_model for cdef in cdefs]
    models = list(set(models))
    for model in models:
        for action in ["view_", "add_", "change_", "delete_", "view_historical"]:
            consent_codenames.append(f".{action}".join(model.split(".")))


consent_codenames.extend(navbar_codenames)
consent_codenames.sort()



from .consent_modelform_mixin import ConsentModelFormMixin
from .requires_consent_modelform_mixin import RequiresConsentModelFormMixin

__all__ = ["ConsentModelFormMixin", "RequiresConsentModelFormMixin"]



from dateutil.relativedelta import relativedelta
from django.test import TestCase, override_settings
from edc_protocol.research_protocol_config import ResearchProtocolConfig
from edc_utils import get_utcnow

from edc_consent.consent_definition import ConsentDefinition
from edc_consent.exceptions import SiteConsentError
from edc_consent.site_consents import site_consents


@override_settings(
    EDC_PROTOCOL_STUDY_OPEN_DATETIME=get_utcnow() - relativedelta(years=5),
    EDC_PROTOCOL_STUDY_CLOSE_DATETIME=get_utcnow() + relativedelta(years=1),
    EDC_AUTH_SKIP_SITE_AUTHS=True,
    EDC_AUTH_SKIP_AUTH_UPDATER=False,
)
class TestConsentModel(TestCase):
    def setUp(self):
        self.study_open_datetime = ResearchProtocolConfig().study_open_datetime
        self.study_close_datetime = ResearchProtocolConfig().study_close_datetime
        site_consents.registry = {}

    def default_options(self, **kwargs):
        options = dict(
            start=self.study_open_datetime,
            end=self.study_close_datetime,
            gender=["M", "F"],
            version="1",
            age_min=16,
            age_max=64,
            age_is_adult=18,
        )
        options.update(**kwargs)
        return options

    def test_ok(self):
        ConsentDefinition("consent_app.subjectconsentv1", **self.default_options())

    def test_cdef_name(self):
        cdef1 = ConsentDefinition("consent_app.subjectconsentv1", **self.default_options())
        self.assertEqual(cdef1.name, "consent_app.subjectconsentv1-1")
        site_consents.register(cdef1)
        site_consents.get_consent_definition("consent_app.subjectconsentv1")
        site_consents.get_consent_definition(model="consent_app.subjectconsentv1")
        site_consents.get_consent_definition(version="1")

        # add country
        site_consents.registry = {}
        cdef1 = ConsentDefinition(
            "consent_app.subjectconsentugv1", **self.default_options(country="uganda")
        )
        self.assertEqual(cdef1.name, "consent_app.subjectconsentugv1-1")
        site_consents.register(cdef1)
        cdef2 = site_consents.get_consent_definition(country="uganda")
        self.assertEqual(cdef1, cdef2)

    def test_with_country(self):
        site_consents.registry = {}
        cdef1 = ConsentDefinition(
            "consent_app.subjectconsentv1", country="uganda", **self.default_options()
        )
        site_consents.register(cdef1)
        cdef2 = site_consents.get_consent_definition(country="uganda")
        self.assertEqual(cdef1, cdef2)

    def test_with_country_raises_on_potential_duplicate(self):
        site_consents.registry = {}
        cdef1 = ConsentDefinition(
            "consent_app.subjectconsentv1", country="uganda", **self.default_options()
        )
        cdef2 = ConsentDefinition(
            "consent_app.subjectconsentugv1", country="uganda", **self.default_options()
        )
        site_consents.register(cdef1)
        site_consents.register(cdef2)
        self.assertRaises(
            SiteConsentError, site_consents.get_consent_definition, country="uganda"
        )

    def test_duplicate_version(self):
        site_consents.registry = {}
        cdef1 = ConsentDefinition(
            "consent_app.subjectconsentv1", country="uganda", **self.default_options()
        )
        cdef2 = ConsentDefinition(
            "consent_app.subjectconsentugv1", country="uganda", **self.default_options()
        )
        site_consents.register(cdef1)
        site_consents.register(cdef2)
        self.assertRaises(
            SiteConsentError, site_consents.get_consent_definition, country="uganda"
        )



from datetime import datetime, timedelta
from zoneinfo import ZoneInfo

import time_machine
from dateutil.relativedelta import relativedelta
from django.conf import settings
from django.contrib.sites.models import Site
from django.test import TestCase, override_settings
from edc_protocol.research_protocol_config import ResearchProtocolConfig
from edc_utils import get_utcnow
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from faker import Faker
from model_bakery import baker

from consent_app.models import CrfOne, SubjectConsent, SubjectVisit
from consent_app.visit_schedules import get_visit_schedule
from edc_consent.field_mixins import IdentityFieldsMixinError
from edc_consent.site_consents import site_consents

from ...exceptions import (
    ConsentDefinitionDoesNotExist,
    ConsentDefinitionModelError,
    NotConsentedError,
)
from ..consent_test_utils import consent_factory

fake = Faker()


@time_machine.travel(datetime(2019, 4, 1, 8, 00, tzinfo=ZoneInfo("UTC")))
@override_settings(
    EDC_PROTOCOL_STUDY_OPEN_DATETIME=get_utcnow() - relativedelta(years=5),
    EDC_PROTOCOL_STUDY_CLOSE_DATETIME=get_utcnow() + relativedelta(years=1),
    EDC_AUTH_SKIP_SITE_AUTHS=True,
    EDC_AUTH_SKIP_AUTH_UPDATER=False,
)
class TestConsentModel(TestCase):
    def setUp(self):
        self.study_open_datetime = ResearchProtocolConfig().study_open_datetime
        self.study_close_datetime = ResearchProtocolConfig().study_close_datetime
        site_consents.registry = {}
        self.consent_v1 = consent_factory(
            proxy_model="consent_app.subjectconsentv1",
            start=self.study_open_datetime,
            end=self.study_open_datetime + timedelta(days=50),
            version="1.0",
        )
        self.consent_v2 = consent_factory(
            proxy_model="consent_app.subjectconsentv2",
            start=self.study_open_datetime + timedelta(days=51),
            end=self.study_open_datetime + timedelta(days=100),
            version="2.0",
        )

        self.consent_v3_start_date = self.study_open_datetime + timedelta(days=101)
        self.consent_v3 = consent_factory(
            proxy_model="consent_app.subjectconsentv3",
            start=self.study_open_datetime + timedelta(days=101),
            end=self.study_open_datetime + timedelta(days=150),
            version="3.0",
            updates=self.consent_v2,
        )
        site_consents.register(self.consent_v1)
        site_consents.register(self.consent_v2, updated_by=self.consent_v3)
        site_consents.register(self.consent_v3)
        self.dob = self.study_open_datetime - relativedelta(years=25)

    def test_encryption(self):
        subject_consent = baker.make_recipe(
            "consent_app.subjectconsentv1",
            first_name="ERIK",
            consent_datetime=self.study_open_datetime,
            dob=get_utcnow() - relativedelta(years=25),
        )
        self.assertEqual(subject_consent.first_name, "ERIK")

    def test_gets_subject_identifier(self):
        """Asserts a blank subject identifier is set to the
        subject_identifier_as_pk.
        """
        consent = baker.make_recipe(
            "consent_app.subjectconsentv1",
            subject_identifier=None,
            consent_datetime=self.study_open_datetime,
            dob=get_utcnow() - relativedelta(years=25),
            site=Site.objects.get_current(),
        )
        self.assertIsNotNone(consent.subject_identifier)
        self.assertNotEqual(consent.subject_identifier, consent.subject_identifier_as_pk)
        consent.save()
        self.assertIsNotNone(consent.subject_identifier)
        self.assertNotEqual(consent.subject_identifier, consent.subject_identifier_as_pk)

    def test_subject_has_current_consent(self):
        subject_identifier = "123456789"
        identity = "987654321"
        baker.make_recipe(
            "consent_app.subjectconsentv1",
            subject_identifier=subject_identifier,
            identity=identity,
            confirm_identity=identity,
            consent_datetime=self.study_open_datetime + timedelta(days=1),
            dob=get_utcnow() - relativedelta(years=25),
        )
        subject_consent = site_consents.get_consent_or_raise(
            subject_identifier="123456789",
            report_datetime=self.study_open_datetime + timedelta(days=1),
            site_id=settings.SITE_ID,
        )
        self.assertEqual(subject_consent.version, "1.0")
        baker.make_recipe(
            "consent_app.subjectconsentv2",
            subject_identifier=subject_identifier,
            identity=identity,
            confirm_identity=identity,
            consent_datetime=self.study_open_datetime + timedelta(days=60),
            dob=get_utcnow() - relativedelta(years=25),
        )
        subject_consent = site_consents.get_consent_or_raise(
            subject_identifier="123456789",
            report_datetime=self.study_open_datetime + timedelta(days=60),
            site_id=settings.SITE_ID,
        )
        self.assertEqual(subject_consent.version, "2.0")

    def test_model_updates_version_according_to_cdef_used(self):
        """Asserts the consent model finds the cdef and updates
        column `version` using to the version number on the
        cdef.
        """
        subject_identifier = "123456789"
        identity = "987654321"
        consent = baker.make_recipe(
            "consent_app.subjectconsentv1",
            subject_identifier=subject_identifier,
            identity=identity,
            confirm_identity=identity,
            consent_datetime=self.study_open_datetime,
            dob=get_utcnow() - relativedelta(years=25),
        )
        self.assertEqual(consent.version, "1.0")
        consent = baker.make_recipe(
            "consent_app.subjectconsentv2",
            subject_identifier=subject_identifier,
            identity=identity,
            confirm_identity=identity,
            consent_datetime=self.study_open_datetime + timedelta(days=51),
            dob=get_utcnow() - relativedelta(years=25),
        )
        self.assertEqual(consent.version, "2.0")
        consent = baker.make_recipe(
            "consent_app.subjectconsentv3",
            subject_identifier=subject_identifier,
            identity=identity,
            confirm_identity=identity,
            consent_datetime=self.study_open_datetime + timedelta(days=101),
            dob=get_utcnow() - relativedelta(years=25),
        )
        self.assertEqual(consent.version, "3.0")

    def test_model_updates_version_according_to_cdef_used2(self):
        """Asserts the consent model finds the `cdef` and updates
        column `version` using to the version number on the
        `cdef`.

        Note: we get the `model_cls` by looking up the `cdef` first.
        """
        subject_identifier = "123456789"
        identity = "987654321"
        cdef = site_consents.get_consent_definition(report_datetime=self.study_open_datetime)
        consent = baker.make_recipe(
            cdef.model,
            subject_identifier=subject_identifier,
            identity=identity,
            confirm_identity=identity,
            consent_datetime=self.study_open_datetime,
            dob=get_utcnow() - relativedelta(years=25),
        )
        self.assertEqual(consent.version, "1.0")
        cdef = site_consents.get_consent_definition(report_datetime=self.study_open_datetime)
        self.assertRaises(
            ConsentDefinitionModelError,
            baker.make_recipe,
            cdef.model,
            subject_identifier=subject_identifier,
            identity=identity,
            confirm_identity=identity,
            consent_datetime=self.study_open_datetime + timedelta(days=101),
            dob=get_utcnow() - relativedelta(years=25),
        )

        cdef = site_consents.get_consent_definition(
            report_datetime=self.study_open_datetime + timedelta(days=101)
        )
        consent = baker.make_recipe(
            cdef.model,
            subject_identifier=subject_identifier,
            identity=identity,
            confirm_identity=identity,
            consent_datetime=self.study_open_datetime + timedelta(days=101),
            dob=get_utcnow() - relativedelta(years=25),
        )
        self.assertEqual(consent.version, "3.0")

    def test_model_correctly_gets_v3_by_date(self):
        """Asserts that a consent model instance created when the
        current date is within the V3 validity period correctly
        has `instance.version == 3.0`.
        """
        traveller = time_machine.travel(self.study_open_datetime + timedelta(days=110))
        traveller.start()
        subject_identifier = "123456789"
        identity = "987654321"
        consent = baker.make_recipe(
            "consent_app.subjectconsentv3",
            subject_identifier=subject_identifier,
            identity=identity,
            confirm_identity=identity,
            consent_datetime=get_utcnow(),
            dob=get_utcnow() - relativedelta(years=25),
        )
        self.assertEqual(consent.version, "3.0")

    def test_model_updates_from_v1_to_v2(self):
        """Assert, for a single participant, a second consent model
        instance submitted within the v2 validity period has
        version == 2.0.

        Also note that there are now 2 instances of the consent
        model for this participant.
        """
        subject_identifier = "123456789"
        identity = "987654321"

        # travel to V1 validity period
        traveller = time_machine.travel(self.study_open_datetime)
        traveller.start()
        cdef = site_consents.get_consent_definition(report_datetime=get_utcnow())
        subject_consent = baker.make_recipe(
            cdef.model,
            subject_identifier=subject_identifier,
            identity=identity,
            confirm_identity=identity,
            consent_datetime=get_utcnow(),
            dob=get_utcnow() - relativedelta(years=25),
        )
        self.assertEqual(subject_consent.version, "1.0")
        self.assertEqual(subject_consent.subject_identifier, subject_identifier)
        self.assertEqual(subject_consent.identity, identity)
        self.assertEqual(subject_consent.confirm_identity, identity)
        self.assertEqual(subject_consent.version, cdef.version)
        self.assertEqual(subject_consent.consent_definition_name, cdef.name)
        traveller.stop()

        # travel to V2 validity period
        # create second consent for the same individual
        traveller = time_machine.travel(self.study_open_datetime + timedelta(days=51))
        traveller.start()
        cdef = site_consents.get_consent_definition(report_datetime=get_utcnow())
        subject_consent = cdef.model_cls(
            subject_identifier=subject_identifier,
            identity=identity,
            confirm_identity=identity,
            consent_datetime=get_utcnow(),
            dob=get_utcnow() - relativedelta(years=25),
        )
        subject_consent.save()
        subject_consent.refresh_from_db()
        self.assertEqual(subject_consent.version, "2.0")
        self.assertEqual(subject_consent.subject_identifier, subject_identifier)
        self.assertEqual(subject_consent.identity, identity)
        self.assertEqual(subject_consent.confirm_identity, identity)
        self.assertEqual(subject_consent.consent_definition_name, cdef.name)

        self.assertEqual(SubjectConsent.objects.filter(identity=identity).count(), 2)

    def test_first_consent_is_v2(self):
        traveller = time_machine.travel(self.study_open_datetime + timedelta(days=51))
        traveller.start()
        subject_identifier = "123456789"
        identity = "987654321"

        cdef = site_consents.get_consent_definition(report_datetime=get_utcnow())
        self.assertEqual(cdef.version, "2.0")
        subject_consent = baker.make_recipe(
            cdef.model,
            subject_identifier=subject_identifier,
            identity=identity,
            confirm_identity=identity,
            consent_datetime=get_utcnow(),
            dob=get_utcnow() - relativedelta(years=25),
        )
        self.assertEqual(subject_consent.subject_identifier, subject_identifier)
        self.assertEqual(subject_consent.identity, identity)
        self.assertEqual(subject_consent.confirm_identity, identity)
        self.assertEqual(subject_consent.version, cdef.version)
        self.assertEqual(subject_consent.consent_definition_name, cdef.name)

    def test_first_consent_is_v3(self):
        traveller = time_machine.travel(self.study_open_datetime + timedelta(days=101))
        traveller.start()
        subject_identifier = "123456789"
        identity = "987654321"

        cdef = site_consents.get_consent_definition(report_datetime=get_utcnow())
        self.assertEqual(cdef.version, "3.0")
        subject_consent = baker.make_recipe(
            cdef.model,
            subject_identifier=subject_identifier,
            identity=identity,
            confirm_identity=identity,
            consent_datetime=get_utcnow(),
            dob=get_utcnow() - relativedelta(years=25),
        )
        self.assertEqual(subject_consent.subject_identifier, subject_identifier)
        self.assertEqual(subject_consent.identity, identity)
        self.assertEqual(subject_consent.confirm_identity, identity)
        self.assertEqual(subject_consent.version, cdef.version)
        self.assertEqual(subject_consent.consent_definition_name, cdef.name)

    def test_raise_with_date_past_any_consent_period(self):
        traveller = time_machine.travel(self.study_open_datetime + timedelta(days=200))
        traveller.start()
        subject_identifier = "123456789"
        identity = "987654321"
        self.assertRaises(
            ConsentDefinitionDoesNotExist,
            site_consents.get_consent_definition,
            report_datetime=get_utcnow(),
        )
        self.assertRaises(
            ConsentDefinitionDoesNotExist,
            baker.make_recipe,
            "consent_app.subjectconsentv1",
            subject_identifier=subject_identifier,
            identity=identity,
            confirm_identity=identity,
            consent_datetime=get_utcnow(),
            dob=get_utcnow() - relativedelta(years=25),
        )

    def test_saving_with_date_past_any_consent_period_without_consent_raises(self):
        subject_identifier = "123456789"
        identity = "987654321"

        datetime_within_consent_v1 = self.study_open_datetime + timedelta(days=10)
        cdef_v1 = site_consents.get_consent_definition(
            report_datetime=datetime_within_consent_v1
        )
        datetime_within_consent_v2 = self.study_open_datetime + timedelta(days=60)
        cdef_v2 = site_consents.get_consent_definition(
            report_datetime=datetime_within_consent_v2
        )
        datetime_within_consent_v3 = self.study_open_datetime + timedelta(days=110)
        cdef_v3 = site_consents.get_consent_definition(
            report_datetime=datetime_within_consent_v3
        )

        visit_schedule = get_visit_schedule([cdef_v1, cdef_v2, cdef_v3])
        schedule = visit_schedule.schedules.get("schedule1")
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule)

        # jump to and test timepoint within consent v1 window
        traveller = time_machine.travel(datetime_within_consent_v1)
        traveller.start()

        # try subject visit before consenting
        self.assertRaises(
            NotConsentedError,
            SubjectVisit.objects.create,
            report_datetime=get_utcnow(),
            subject_identifier=subject_identifier,
            visit_schedule_name=visit_schedule.name,
            schedule_name=schedule.name,
        )

        # consent and try again
        subject_consent = baker.make_recipe(
            cdef_v1.model,
            subject_identifier=subject_identifier,
            identity=identity,
            confirm_identity=identity,
            consent_datetime=get_utcnow(),
            dob=get_utcnow() - relativedelta(years=25),
        )
        self.assertEqual(subject_consent.consent_definition_name, cdef_v1.name)
        self.assertEqual(subject_consent.version, "1.0")
        self.assertEqual(cdef_v1.model, "consent_app.subjectconsentv1")

        try:
            subject_visit = SubjectVisit.objects.create(
                report_datetime=get_utcnow(),
                subject_identifier=subject_identifier,
                visit_schedule_name=visit_schedule.name,
                schedule_name=schedule.name,
            )
            subject_visit.save()
            crf_one = CrfOne.objects.create(
                subject_visit=subject_visit,
                subject_identifier=subject_identifier,
                report_datetime=get_utcnow(),
            )
            crf_one.save()
        except NotConsentedError:
            self.fail("NotConsentedError unexpectedly raised")
        traveller.stop()

        # jump to and test timepoint within consent v2 window
        traveller = time_machine.travel(datetime_within_consent_v2)
        traveller.start()

        # try subject visit before consenting (v2)
        self.assertRaises(
            NotConsentedError,
            SubjectVisit.objects.create,
            report_datetime=get_utcnow(),
            subject_identifier=subject_identifier,
            visit_schedule_name=visit_schedule.name,
            schedule_name=schedule.name,
        )

        # consent (v2) and try again
        subject_consent = baker.make_recipe(
            cdef_v2.model,
            subject_identifier=subject_identifier,
            identity=identity,
            confirm_identity=identity,
            consent_datetime=get_utcnow(),
            dob=get_utcnow() - relativedelta(years=25),
        )
        self.assertEqual(subject_consent.consent_definition_name, cdef_v2.name)
        self.assertEqual(subject_consent.version, "2.0")
        self.assertEqual(cdef_v2.model, "consent_app.subjectconsentv2")

        try:
            subject_visit = SubjectVisit.objects.create(
                report_datetime=get_utcnow(),
                subject_identifier=subject_identifier,
                visit_schedule_name=visit_schedule.name,
                schedule_name=schedule.name,
            )
            subject_visit.save()
            crf_one = CrfOne.objects.create(
                subject_visit=subject_visit,
                subject_identifier=subject_identifier,
                report_datetime=get_utcnow(),
            )
            crf_one.save()
        except NotConsentedError:
            self.fail("NotConsentedError unexpectedly raised")
        traveller.stop()

        # jump to and test timepoint within consent v3 window
        traveller = time_machine.travel(datetime_within_consent_v3)
        traveller.start()

        # try subject visit before consenting (v3)
        self.assertRaises(
            NotConsentedError,
            SubjectVisit.objects.create,
            report_datetime=get_utcnow(),
            subject_identifier=subject_identifier,
            visit_schedule_name=visit_schedule.name,
            schedule_name=schedule.name,
        )

        # consent (v3) and try again
        subject_consent = baker.make_recipe(
            cdef_v3.model,
            subject_identifier=subject_identifier,
            identity=identity,
            confirm_identity=identity,
            consent_datetime=get_utcnow(),
            dob=get_utcnow() - relativedelta(years=25),
        )
        self.assertEqual(subject_consent.consent_definition_name, cdef_v3.name)
        self.assertEqual(subject_consent.version, "3.0")
        self.assertEqual(cdef_v3.model, "consent_app.subjectconsentv3")

        try:
            subject_visit = SubjectVisit.objects.create(
                report_datetime=get_utcnow(),
                subject_identifier=subject_identifier,
                visit_schedule_name=visit_schedule.name,
                schedule_name=schedule.name,
            )
            subject_visit.save()
            crf_one = CrfOne.objects.create(
                subject_visit=subject_visit,
                subject_identifier=subject_identifier,
                report_datetime=get_utcnow(),
            )
            crf_one.save()
        except NotConsentedError:
            self.fail("NotConsentedError unexpectedly raised")
        traveller.stop()

    def test_save_crf_with_consent_end_shortened_to_before_existing_subject_visit_raises(
        self,
    ):

        traveller = time_machine.travel(self.study_open_datetime)
        traveller.start()
        subject_identifier = "123456789"
        identity = "987654321"

        cdef_v1 = site_consents.get_consent_definition(
            report_datetime=self.study_open_datetime + timedelta(days=10)
        )
        cdef_v2 = site_consents.get_consent_definition(
            report_datetime=self.study_open_datetime + timedelta(days=60)
        )
        datetime_within_consent_v3 = self.study_open_datetime + timedelta(days=110)
        cdef_v3 = site_consents.get_consent_definition(
            report_datetime=datetime_within_consent_v3
        )

        visit_schedule = get_visit_schedule([cdef_v1, cdef_v2, cdef_v3])
        schedule = visit_schedule.schedules.get("schedule1")
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule)

        traveller.stop()
        traveller = time_machine.travel(datetime_within_consent_v3)
        traveller.start()

        # consent v3
        subject_consent = baker.make_recipe(
            cdef_v3.model,
            subject_identifier=subject_identifier,
            identity=identity,
            confirm_identity=identity,
            consent_datetime=get_utcnow(),
            dob=get_utcnow() - relativedelta(years=25),
        )
        self.assertEqual(subject_consent.consent_definition_name, cdef_v3.name)
        self.assertEqual(subject_consent.version, "3.0")
        self.assertEqual(cdef_v3.model, "consent_app.subjectconsentv3")

        # create two visits within consent v3 period
        subject_visit_1 = SubjectVisit.objects.create(
            report_datetime=get_utcnow(),
            subject_identifier=subject_identifier,
            visit_schedule_name=visit_schedule.name,
            schedule_name=schedule.name,
        )
        subject_visit_1.save()
        # subject_visit_2 = SubjectVisit.objects.create(
        #     report_datetime=get_utcnow() + relativedelta(days=20),
        #     subject_identifier=subject_identifier,
        #     visit_schedule_name=visit_schedule.name,
        #     schedule_name=schedule.name,
        # )
        # subject_visit_2.save()

        # cut short v3 validity period, and introduce new v4 consent definition,
        updated_v3_end_datetime = datetime_within_consent_v3 + relativedelta(days=1)
        site_consents.registry[cdef_v3.name].end = updated_v3_end_datetime
        site_consents.registry[cdef_v3.name].updated_by = "4.0"
        self.assertEqual(site_consents.registry[cdef_v3.name].end, updated_v3_end_datetime)
        self.assertEqual(site_consents.registry[cdef_v3.name].end, cdef_v3.end)
        self.assertEqual(site_consents.registry[cdef_v3.name].updated_by, "4.0")
        self.assertEqual(site_consents.registry[cdef_v3.name].updated_by, cdef_v3.updated_by)

        consent_v4 = consent_factory(
            proxy_model="consent_app.subjectconsentv4",
            start=cdef_v3.end + relativedelta(days=1),
            end=self.study_open_datetime + timedelta(days=150),
            version="4.0",
            updates=self.consent_v3,
        )

        site_consents.unregister(self.consent_v3)
        site_consents.register(self.consent_v3, updated_by=consent_v4)
        site_consents.register(consent_v4)

        traveller.stop()
        traveller = time_machine.travel(cdef_v3.end + relativedelta(days=20))
        traveller.start()
        cdef_v4 = site_consents.get_consent_definition(report_datetime=get_utcnow())
        self.assertEqual(cdef_v4.version, "4.0")
        schedule.consent_definitions = [cdef_v1, cdef_v2, cdef_v3, cdef_v4]

        # try saving CRF within already consented (v3) period
        try:
            crf_one = CrfOne.objects.create(
                subject_visit=subject_visit_1,
                subject_identifier=subject_identifier,
                report_datetime=datetime_within_consent_v3,
            )
        except NotConsentedError:
            self.fail("NotConsentedError unexpectedly raised")
        try:
            crf_one.save()
        except NotConsentedError:
            self.fail("NotConsentedError unexpectedly raised")

        # now try to save CRF at within v4 period
        crf_one.report_datetime = get_utcnow()
        self.assertRaises(NotConsentedError, crf_one.save)

        # consent v4 and try again
        subject_consent = baker.make_recipe(
            cdef_v4.model,
            subject_identifier=subject_identifier,
            identity=identity,
            confirm_identity=identity,
            consent_datetime=get_utcnow(),
            dob=get_utcnow() - relativedelta(years=25),
        )
        self.assertEqual(subject_consent.consent_definition_name, cdef_v4.name)
        self.assertEqual(subject_consent.version, "4.0")
        self.assertEqual(cdef_v4.model, "consent_app.subjectconsentv4")

        try:
            crf_one = CrfOne.objects.create(
                subject_visit=subject_visit_1,
                subject_identifier=subject_identifier,
                report_datetime=get_utcnow(),
            )
            crf_one.save()
        except NotConsentedError:
            self.fail("NotConsentedError unexpectedly raised")
        traveller.stop()

    def test_raise_with_incorrect_model_for_cdef(self):
        traveller = time_machine.travel(self.study_open_datetime + timedelta(days=120))
        traveller.start()
        subject_identifier = "123456789"
        identity = "987654321"
        cdef = site_consents.get_consent_definition(report_datetime=get_utcnow())
        self.assertEqual(cdef.model, "consent_app.subjectconsentv3")
        self.assertRaises(
            ConsentDefinitionModelError,
            baker.make_recipe,
            "consent_app.subjectconsentv1",
            subject_identifier=subject_identifier,
            identity=identity,
            confirm_identity=identity,
            consent_datetime=get_utcnow(),
            dob=get_utcnow() - relativedelta(years=25),
        )

    def test_model_str_repr_etc(self):
        obj = baker.make_recipe(
            "consent_app.subjectconsentv1",
            screening_identifier="ABCDEF",
            subject_identifier="12345",
            consent_datetime=self.study_open_datetime + relativedelta(days=1),
        )

        self.assertTrue(str(obj))
        self.assertTrue(repr(obj))
        self.assertTrue(obj.age_at_consent)
        self.assertTrue(obj.formatted_age_at_consent)
        self.assertEqual(obj.report_datetime, obj.consent_datetime)

    def test_checks_identity_fields_match_or_raises(self):
        self.assertRaises(
            IdentityFieldsMixinError,
            baker.make_recipe,
            "consent_app.subjectconsentv1",
            subject_identifier="12345",
            consent_datetime=self.study_open_datetime + relativedelta(days=1),
            identity="123456789",
            confirm_identity="987654321",
        )



from django.conf import settings
from edc_auth.constants import PII, PII_VIEW
from edc_auth.site_auths import site_auths
from edc_auth.utils import remove_default_model_permissions_from_edc_permissions

from .auth_objects import consent_codenames, navbar_tuples

site_auths.add_post_update_func(
    "edc_consent",
    remove_default_model_permissions_from_edc_permissions,
)

site_auths.add_custom_permissions_tuples(
    model="edc_consent.edcpermissions", codename_tuples=navbar_tuples
)

site_auths.update_group(*consent_codenames, name=PII, no_delete=True)
site_auths.update_group(*consent_codenames, name=PII_VIEW, view_only=True)
site_auths.add_pii_model(settings.SUBJECT_CONSENT_MODEL)



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "consent_app"
    verbose_name = "Edc Consent test app"
    include_in_administration_section = False



from django.db import models


class RequiresConsentFieldsModelMixin(models.Model):
    """See pre-save signal that checks if subject is consented"""

    consent_model = models.CharField(max_length=50, null=True, blank=True)

    consent_version = models.CharField(max_length=10, null=True, blank=True)

    class Meta:
        abstract = True
